% demo_history.m is a demonstration program illustrating a feature 
% that allows efficient calculation of the accuracy of all low rank
% approximations less than or equal to a specified rank.
%
% To execute type: 
%    demo_history

% Written by L. Foster, 2008-04-16
% (C) copyright 2008, Leslie Foster, 2008-04-16.

format compact

disp(' ')
disp('This demonstration uses bootstrap resampling to compare the numerical')
disp('stability, accuracy and efficiency of six algorithms for Gaussian ')
disp('process regression using low-rank approximation.') 
disp(' ')
disp('The demonstration also illustrates the ability to calculate the ')
disp('accuracy of the predictions for all low rank approximations ')
disp('less than a specified rank. This can be helpful in trying to ')
disp('identify an appropriate rank for a low rank approximation ')
disp('technique.  The algorithms are:')
disp(' ')
disp('   gprSRPPn  - the subset of regressors method, using an algorithm ')
disp('       which solves normal equations with the active set')
disp('       selected randomly')
disp('   gprSRPPnp - the subset of regressors method, using an algorithm')
disp('       which solves normal equations with an active set selected')
disp('       by  partial Cholesky factorization with pivoting.')
disp('   gprSRPPv  - the subset of regressors method, using an algorithm')
disp('       based on the V method with the active set selected randomly')
disp('   gprSRPPvp - the subset of regressors method, using an algorithm')
disp('       based on the V method with an active set selected by partial')
disp('       Cholesky with pivoting.')
disp('   gprSRPPq  - the subset of regressors method, using an algorithm')
disp('       based on a QR factorization with the active set selected')
disp('       randomly')
disp('   gprSRPPqp - the subset of regressors method, using an algorithm')
disp('       based on a QR factorization with an active set selected by')
disp('       partial Cholesky factorization with pivoting.')

disp(' ')
disp('Press any key to continue')
pause;
disp(' ')

%
disp('We use data for estimation of redshift of galaxies from broad spectrum ')
disp('measurements.  This data is discussed in Way, M.J. and Srivastava, ') 
disp('A.N., The Astrophysical Journal. 647: 102, 2006 and comes from the')
disp('Sloan Digital Sky Survey: York, D.G., et al., The Astrophysical')
disp('Journal, 120: 1579, 2000.')
disp(' ')
disp('The training and testing data is contained in the file redshift_data.mat')
disp('The raw training data is in the input matrix X (180045 x 5) and the ')
disp('vector y (180045 x 1).  The test data is contained in the matrix Xtest ')
disp('(20229 x 5 ) and vector ytest (20229 x 1). ')
disp(' ')
disp('  load redshift_data.mat')
disp(' ')
load redshift_data.mat

[n,D]= size(X);

if ( 1 == 1 )
    % this alternative uses part of the data and runs relatively quickly
    nsubset = 10000;       % size of subset of the data for each resampling run   
    m_rank = 100;         % rank of low rank approximation (size of active set)
    nsamples = 10;                            % number of samples in resampling
    %nsamples = 100;           % nsamples = 100 may be more reliable but slower
    replacement = 0;          % use resampling without replacement in selecting
                                                               % subset of data
elseif ( 1 == 0 )
    % this alternative uses all the data and runs somewhat slower
    nsubset = n;                     % the subset of the data is the entire set  
    m_rank = 100;         % rank of low rank approximation (size of active set)
    nsamples = 10;                            % number of samples in resampling
    %nsamples = 100;           % nsamples = 100 may be more reliable but slower
    replacement = 1;             % use resampling with replacement in selecting
                                                               % subset of data
elseif ( 1 == 0 )
    % This option produces a figure similar to Figure 3 in
    % the manuscript "Stable and Efficient Gaussian Process Calculations"
    % by L. Foster, et. al., submitted to the  Journal of Machine Learning
    % Research.
    nsubset = round(.5*n); % size of subset of the data for each resampling run
    m_rank = 500;         % rank of low rank approximation (size of active set)
    nsamples =  10;                           % number of samples in resampling
    %nsamples = 100;           % nsamples = 100 may be more reliable but slower
    replacement = 0;          % use resampling without replacement in selecting
                                                               % subset of data
end

disp('The demonstration will use bootstrap resampling.  To make ')
disp('the demonstration run in a reasonable time, for each resampling ')
disp(['run we will select a random subset consisting of ',  ... 
    int2str(nsubset),' points or '])
disp('galaxies from the 180045 galaxies in the training set. Since the sample ')
if ( nsubset < n )
   disp(['size (',int2str(nsubset), ...
    ') is smaller than the size of entire data set we will use '])
   disp('resampling without replacement to choose the sample for each run.')
else
   disp(['size (',int2str(nsubset), ...
    ') equals the size of the entire data set we will use bootstrap '])
   disp('resampling with replacement to choose the sample for each run.')
end
disp(' ')
disp('Also to make the demonstration run in a reasonable time the number ')
disp(['of samples in the resampling is set to ', ...
    int2str(nsamples), ' for the'])
disp('demonstration.')
   
disp(' ')
disp('Press any key to continue')
pause;
disp(' ')

%disp(' ')
disp(['We will use low rank approximations of maximum rank ', ...
    int2str(m_rank), '.'])
disp(['In some cases an algorithm will not work for rank ', ...
    int2str(m_rank)])
disp(['and the maximum rank is reduced.  For each bootstrap run'])
disp(['the ' , ...
    ' elements in the active set will be chosen randomly'])
disp('for the methods without pivoting and will be chosen by')
disp('a partial Cholesky factorization for the methods with pivoting.')

if ( 1 == 1 )
   % alternative to select neural network covariance function
   disp(' ')
   disp('We use Gaussian process regression with a neural network covariance ')
   disp('function (equation 4.29 of Ramussen and Williams, 2006).  The')
   disp('covariance function used is made up of the sum of the neural network')
   disp('covariance function and independent noise. Thus, it is specified as ')
   disp('follows:')
   disp(' ')
   disp('   covfunc = {''covSum'',{''covNNone'',''covNoise''}};')
   disp(' ')
   covfunc = {'covSum',{'covNNone','covNoise'}};
   %
   % NOTE: code for most covariance functions, other than polynomial
   % covariance functions, will be similar to the above code
   % where covNNone would be replaced by the desire covariance function.
end   

if ( 1 == 0 )
   % alternative that chooses hyperparameters using Rasmussen and Williams
   %   minimize and the first hyper_size data points
   hyper_size = m_rank;       % size of data set used to select hyperparameters
   disp('For simplicity in the resampling we select one set ')
   disp('of hyperparameters for all the resampling samples.')
   disp('To do this we initialize the hyperparameters to be all one')
   disp('(so that their logarithms are zero).  Using these initial ')
   disp('values as a starting point the hyperparameters are then trained')
   disp('using the minimize function from Rasmussen and Williams''')
   disp('code.  This will train the hyperparameters by maximizing the ')
   disp('approximate marginal likelihood of the subset of data method ')
   disp('(see eq. 8.31 or Rasmussen and Williams text).  For simplicity')
   disp('the subset of data that we select in this calculation is the ')
   disp(['first ',int2str(hyper_size),' galaxies in X and y.'])
   disp(' ')
   disp('Press any key to train the hyperparameters:')
   pause
   disp(' ')

   logtheta0 = zeros(eval(feval(covfunc{:})),1);
   INDEX_hyper = 1:hyper_size;
   logtheta = ...
      minimize(logtheta0,'gpr',-100,covfunc,X(INDEX_hyper,:),y(INDEX_hyper));
   disp('After training the hyperparameters are :')
   disp(num2str(exp(logtheta')))
   disp(' ')   
end

disp('In the resampling runs for each sample we print out run')
disp('times and accuracies of each algorithm and we plot the ')
disp('accuracies of the algorithms for each low rank approximation ')
disp(['with rank less than or equal to ',int2str(m_rank), ...
    '. The accuracy'])
disp('is  measured by the root mean square error (rmse) between')
disp('the predicted redshifts using the algorithm and the measured')
disp('redshift data for the test set.')
disp(' ')
disp('Press any key to start the resampling runs:')
pause
disp(' ')

%Initialize:
rmse = 0;
t_n = zeros(1,nsamples);
rank_n = zeros(1,nsamples);
rmse_n = zeros(nsamples,m_rank);
t_np = zeros(1,nsamples);
rank_np = zeros(1,nsamples);
rmse_np = zeros(nsamples,m_rank);
t_v = zeros(1,nsamples);
rank_v = zeros(1,nsamples);
rmse_v = zeros(nsamples,m_rank);
t_vp = zeros(1,nsamples);
rank_vp = zeros(1,nsamples);
rmse_vp = zeros(nsamples,m_rank);
t_q = zeros(1,nsamples);
rank_q = zeros(1,nsamples);
rmse_q = zeros(nsamples,m_rank);
t_qp = zeros(1,nsamples);
rank_qp = zeros(1,nsamples);
rmse_qp = zeros(nsamples,m_rank);

for i = 1:nsamples
    
    % choose a random training set of size m
    if (replacement == 1 )
        % use bootstrap resampling with replacement
        INDEX_s = ceil( n * rand(nsubset,1) );
    else
        % use bootstrap resampling without replacement
        perm = randperm(n);
        INDEX_s = perm(1:nsubset);
    end
    Xs = X(INDEX_s,:);
    ys = y(INDEX_s,:);
    
    % choose a random set of m_rank indices for the active set
    perm = randperm(nsubset);
    INDEX_a = perm(1:m_rank);

    if ( 1 == 1 )
       % alternative that trains the hyperparameters using the
       % active set selected in each  resampling sample
       logtheta0 = zeros(eval(feval(covfunc{:})),1);
       INDEX_hyper = INDEX_a;
       disp(' ')
       disp('train the hyperparameters using the current active set:')
       disp(' ')
       logtheta = minimize(logtheta0,'gpr',-100,covfunc,X(INDEX_hyper,:), ...
           y(INDEX_hyper));
       disp('The hyperparameters are :')
       disp(num2str(exp(logtheta')))
       %disp(' ')
    end
 
    disp(' ')
    disp(['Sample ',int2str(i),' of ',int2str(nsamples)])
   
    % gprSRPPn - normal method without pivoting,   history
    disp(['Running gprSRPPn  which uses the normal equations '])
    tic
    [mu,rmse] = gprSRPPn(logtheta, covfunc, Xs, INDEX_a, ys, Xtest,ytest);
    t_n(i) = toc;                                               % save the time
    rank_calc = length(rmse);   % the actual rank of the low rank approximation
    rank_n(i) = rank_calc;                                    %  save this rank
    rmse_n(i,1:rank_calc) = rmse;                      % save the error history
    if ( rank_calc < m_rank )               % if the actual rank < desired rank
       rmse_n(i,rank_calc:m_rank) = rmse(end);     % pad the error history with 
    end                                             % the last calculated error
    disp(['   rmse = ', num2str(rmse_n(i,end)),' for rank ', ...
        int2str(rank_n(i)),', time = ',num2str(t_n(i))])

    % gprSRPPnp - normal method with pivoting,   history
    disp(['Running gprSRPPnp which uses the normal equations and pivoting'])
    tic
    i_reuse_order = 0;  % set i_reuse_order = 1 to save time for large problems
    if ( i_reuse_order == 0 )
       [mu,rmse] = gprSRPPnp(logtheta, covfunc, Xs, INDEX_a, ys, Xtest,ytest);
    else
       % select this alternative to avoid calculating chol_part for each of 
       % gprSRPPnp, gprSRPPvp and gprSRPPqp in order to save time.
       % Note that gprSRPPnp is equivalent to gprSRPPn preceded by chol_part.
       ipiv = 1;                 % ipiv = 1 forces pivoting in partial Cholesky 
       if ( 1 == 1 )
          tol = 0;  % reduce numerical rank if K becomes semidefinite or indef.
       else
          tol = -1;      % make numerical rank choice based on relative machine 
                    % precision -- see chol_part documentation for more details
       end                               
       [V,rank_cp,piv] =  ...
          chol_part(logtheta,covfunc,Xs,tol,length(INDEX_a),ipiv);
       clear V
       [mu,rmse] = gprSRPPn(logtheta,covfunc,Xs,piv(1:rank_cp),ys,Xtest,ytest);
    end
    t_np(i) = toc;                                              % save the time
    rank_calc = length(rmse);   % the actual rank of the low rank approximation
    rank_np(i) = rank_calc;                                   %  save this rank
    rmse_np(i,1:rank_calc) = rmse;                     % save the error history
    if ( rank_calc < m_rank )               % if the actual rank < desired rank
       rmse_np(i,rank_calc:m_rank) = rmse(end);    % pad the error history with 
    end                                             % the last calculated error
    disp(['   rmse = ', num2str(rmse_np(i,end)),' for rank ', ...
        int2str(rank_np(i)),', time = ',num2str(t_np(i))])

    % gprSRPPv - V method without pivoting,   history
    disp(['Running gprSRPPv  which uses the V method '])
    tic
    [mu,rmse] = gprSRPPv(logtheta, covfunc, Xs, INDEX_a, ys, Xtest,ytest);
    t_v(i) = toc;                                               % save the time
    rank_calc = length(rmse);   % the actual rank of the low rank approximation
    rank_v(i) = rank_calc;                                    %  save this rank
    rmse_v(i,1:rank_calc) = rmse;                      % save the error history
    if ( rank_calc < m_rank )               % if the actual rank < desired rank
       rmse_v(i,rank_calc:m_rank) = rmse(end);     % pad the error history with 
    end                                             % the last calculated error
    disp(['   rmse = ', num2str(rmse_v(i,end)),' for rank ', ...
        int2str(rank_v(i)),', time = ',num2str(t_v(i))])


    % gprSRPPvp - V method with pivoting,   history
    disp(['Running gprSRPPvp which uses the V method and pivoting'])
    tic
    if ( i_reuse_order == 0 )
       [mu,rmse] = gprSRPPvp(logtheta, covfunc, Xs, INDEX_a, ys, Xtest,ytest);
    else
       % select to reuse chol_part ordering piv if it was calculated earlier 
       [mu,rmse] = gprSRPPv(logtheta,covfunc,Xs,piv(1:rank_cp),ys,Xtest,ytest);
    end
    t_vp(i) = toc;                                              % save the time
    rank_calc = length(rmse);   % the actual rank of the low rank approximation
    rank_vp(i) = rank_calc;                                   %  save this rank
    rmse_vp(i,1:rank_calc) = rmse;                     % save the error history
    if ( rank_calc < m_rank )               % if the actual rank < desired rank
       rmse_vp(i,rank_calc:m_rank) = rmse(end);    % pad the error history with 
    end                                             % the last calculated error
    disp(['   rmse = ', num2str(rmse_vp(i,end)),' for rank ', ...
        int2str(rank_vp(i)),', time = ',num2str(t_vp(i))])

    % gprSRPPq - QR factorization without pivoting,   history
    disp(['Running gprSRPPq  which uses a QR factorization '])
    tic
    [mu,rmse] = gprSRPPq(logtheta, covfunc, Xs, INDEX_a, ys, Xtest,ytest);
    t_q(i) = toc;                                               % save the time
    rank_calc = length(rmse);   % the actual rank of the low rank approximation
    rank_q(i) = rank_calc;                                    %  save this rank
    rmse_q(i,1:rank_calc) = rmse;                      % save the error history
    if ( rank_calc < m_rank )               % if the actual rank < desired rank
       rmse_q(i,rank_calc:m_rank) = rmse(end);     % pad the error history with 
    end                                             % the last calculated error
    disp(['   rmse = ', num2str(rmse_q(i,end)),' for rank ', ...
        int2str(rank_q(i)),', time = ',num2str(t_q(i))])

    % gprSRPPqp - QR factorization with pivoting,   history
    disp(['Running gprSRPPqp which uses a QR factorization and pivoting'])
    tic
    if ( i_reuse_order == 0 )
       [mu,rmse] = gprSRPPqp(logtheta, covfunc, Xs, INDEX_a, ys, Xtest,ytest);
    else
       % select to reuse chol_part ordering piv if it was calculated earlier 
       [mu,rmse] = gprSRPPq(logtheta,covfunc,Xs,piv(1:rank_cp),ys,Xtest,ytest);
    end
    t_qp(i) = toc;                                              % save the time
    rank_calc = length(rmse);   % the actual rank of the low rank approximation
    rank_qp(i) = rank_calc;                                   %  save this rank
    rmse_qp(i,1:rank_calc) = rmse;                     % save the error history
    if ( rank_calc < m_rank )               % if the actual rank < desired rank
       rmse_qp(i,rank_calc:m_rank) = rmse(end);    % pad the error history with 
    end                                             % the last calculated error
    disp(['   rmse = ', num2str(rmse_qp(i,end)),' for rank ', ...
        int2str(rank_qp(i)),', time = ',num2str(t_qp(i))])

    
    disp(' ')
    
    if ( 1 == 1 )
       % plot the errors versus rank for each sample
       figure(1)
       clf
       plot(1:m_rank,rmse_n(i,:),'b', ...
            1:m_rank,rmse_np(i,:),'r', ...
            1:m_rank,rmse_v(i,:),'g', ...
            1:m_rank,rmse_vp(i,:),'c', ...
            1:m_rank,rmse_q(i,:),'m', ...
            1:m_rank,rmse_qp(i,:),'k')
     
       xlabel('rank of low rank approximation')
       ylabel('RMS errors for testing data')
       title(['sample ',int2str(i),': RMS errors versus rank'])
       legend('gprSRPPn','gprSRPPnp','gprSRPPv','gprSRPPvp', ...
              'gprSRPPq','qprSRPPqp')
       grid
       shg
       pause(1)  %facilitates viewing results for each sample, delete for speed
    end
   

end


%output summary information

% print results for this sample

disp(' ')
disp(['Summary information for ',int2str(nsamples),' samples'])
disp(' ')
disp(['mean rmse = ', num2str(mean(rmse_n(:,end))), ... 
      ' for gprSRPPn,  mean time = ',num2str(mean(t_n))])
disp(['mean rmse = ', num2str(mean(rmse_np(:,end))), ...
    ' for gprSRPPnp, mean time = ',num2str(mean(t_np))])
disp(['mean rmse = ', num2str(mean(rmse_v(:,end))), ...
    ' for gprSRPPv,  mean time = ',num2str(mean(t_v))])
disp(['mean rmse = ', num2str(mean(rmse_vp(:,end))), ...
    ' for gprSRPPvp, mean time = ',num2str(mean(t_vp))])
disp(['mean rmse = ', num2str(mean(rmse_q(:,end))), ...
    ' for gprSRPPq,  mean time = ',num2str(mean(t_q))])
disp(['mean rmse = ', num2str(mean(rmse_qp(:,end))), ...
    ' for gprSRPPqp, mean time = ',num2str(mean(t_qp))])
    
disp(' ')
disp(['max  rmse = ', num2str(max(rmse_n(:,end))), ... 
      ' for gprSRPPn,  max  time = ',num2str(max(t_n))])
disp(['max  rmse = ', num2str(max(rmse_np(:,end))), ...
    ' for gprSRPPnp, max  time = ',num2str(max(t_np))])
disp(['max  rmse = ', num2str(max(rmse_v(:,end))), ...
    ' for gprSRPPv,  max  time = ',num2str(max(t_v))])
disp(['max  rmse = ', num2str(max(rmse_vp(:,end))), ...
    ' for gprSRPPvp, max  time = ',num2str(max(t_vp))])
disp(['max  rmse = ', num2str(max(rmse_q(:,end))), ...
    ' for gprSRPPq,  max  time = ',num2str(max(t_q))])
disp(['max  rmse = ', num2str(max(rmse_qp(:,end))), ...
    ' for gprSRPPqp, max  time = ',num2str(max(t_qp))])    
        

if ( 1 == 1 )
   % plot the mean errors versus rank
   disp(' ')
   disp('Also see Figure 2.')
   
   figure(2)
   clf   
   plot(1:m_rank,mean(rmse_n),'b', ...
        1:m_rank,mean(rmse_np),'r', ...
        1:m_rank,mean(rmse_v),'g', ...
        1:m_rank,mean(rmse_vp),'c', ...
        1:m_rank,mean(rmse_q),'m', ...
        1:m_rank,mean(rmse_qp),'k')
    
   xlabel('rank of low rank approximation')
   ylabel('mean RMS errors for testing data')
   title(['mean RMS errors (for  ',int2str(nsamples),' samples) versus rank'])
   legend('gprSRPPn','gprSRPPnp','gprSRPPv','gprSRPPvp', ...
          'gprSRPPq','qprSRPPqp')
   grid
   shg
end

disp(' ')
disp('Note that the option which calculates a history of all')
disp('approximations of rank less than or equal to a specified rank ')
disp('requires only a modest amount of additional time beyond the time ')
disp('required by the option which does not calculate this history. ')
disp('The ability to cheaply see the history of the accuracy of a set ')
disp('of low rank approximations can be useful in deciding what rank ')
disp('is appropriate to use.')
disp(' ')

