function [mu,S2SR,S2PP,rmse]=gprSRPPvp(logtheta,covfunc,x,INDEX,y,xstar,ystar);

% gprSRPPvp - Carries out approximate Gaussian process regression prediction
% using the subset of regressors (SR) or projected process approximation (PP)
% with the active set selected by a partial Cholesky decomposition with
% pivoting. This version uses a variation of the Peters-Wilkinson approach
% for solving least squares problems (and partial Cholesky factorization
% with pivoting) to increase the numerical stability of the algorithm .
%
% Usage alternatives:
%
%   mu = gprSRPPvp(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, rmse] = gprSRPPvp(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, S2SR, S2PP] = gprSRPPvp(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, S2SR, S2PP, rmse] = gprSRPPvp(logtheta, covfunc, X, INDEX, y, Xstar);
%   [mu, rmse] = gprSRPPvp(logtheta, covfunc, X, INDEX, y, Xstar, ystar);
%   [mu, S2SR, S2PP,rmse] = gprSRPPvp(logtheta,covfunc,X,INDEX,y,Xstar, ystar);
%
% where
%
% Input (ystar is optional):
%   logtheta a (column) vector of log hyperparameters
%   covfunc  the covariance function, which is assumed to
%            be a covSum, and the last entry of the sum is covNoise
%   x        a n by D matrix of training inputs
%   INDEX    If length(INDEX) == 1 then m = INDEX is the maximum rank of a 
%                 low rank approximation.
%            If length(INDEX) > 1  then m = length(INDEX) is the maximum
%                rank of the low rank approximation. Note that in this
%                case only length(INDEX) is used in the routine.
%            The active set is chosen using a partial Cholesky factorization
%            with pivoting. This factorization selects linearly independent
%            columns of the covariance matrix K for the active set.  If 
%            partial Cholesky factorization with pivoting determines that
%            K has (numerical) rank = rank1 < m, the routine reduces m to
%            rank1.
%   y        a (column) vector (of size n) of targets
%   xstar    a nstar by D matrix of test inputs
%   ystar    is a (column) vector (of size nstar) of test targets
% Output (S2SR, S2PP and rmse are optional):
%   mu       a (column) vector (of size nstar) of predicted means
%   S2SR     a (column) vector (of size nstar) of predicted variances under SR
%   S2PP     a (column) vector (of size nstar) of predicted variances under PP
%   rmse     a (column) vector (of size m) of root mean square errors
%            If the number of input arguments is 6:
%               rmse(i) = || y_est - y || / sqrt(n), i = 1,...,m
%            where y_est is the estimated value of y with a rank i approx.
%            If the number of input arguments is 7:
%               rmse(i) = || y_pred - ystar || / sqrt(nstar), i = 1,...,m
%            where y_pred is the predicted value of ystar with a rank i approx.
%            Note that length(rmse) gives the current (potentially reduced) m.
%
% For more help on covariance functions, see "help covFunctions".

% (C) modification (2008-1-23) by Leslie Foster of code by Chris Williams.

% Notes:    memory and time can be reduced by changing rank_increment =1 below.

if ischar(covfunc), covfunc = cellstr(covfunc); end % convert to cell if needed
[n, D] = size(x);
if eval(feval(covfunc{:})) ~= size(logtheta, 1)
  error('Error: Number of parameters do not agree with covariance function')
end

% we check that the covfunc cell array is a covSum, with last entry 'covNoise'
if length(covfunc) ~= 2 | ~strcmp(covfunc(1), 'covSum') | ...
                                           ~strcmp(covfunc{2}(end), 'covNoise')
  error('The covfunc must be "covSum" whose last summand must be "covNoise"')
end

if ( length(INDEX) > 1 )
   m = length(INDEX);
else
   m = INDEX;
end
if ( m > n )
   m = n;
end

sigma2n = exp(2*logtheta(end));                                % noise variance
[nstar, D] = size(xstar);   % number of test cases and dimension of input space

ipiv = 1;                        % ipiv = 1 forces pivoting in partial cholesky
if ( 1 == 1 )
   tol = 0;     % reduce numerical rank if K becomes semidefinite or indefinite
else
   tol = -1;   % make numerical rank choice based on relative machine precision
                                 % see chol_part documentation for more details
end                               

[V,rank1,piv] =  ...
      chol_part(logtheta,covfunc,x,tol,m,ipiv);           % partial cholesky
if ( rank1 < m )
   m = rank1;                      % redefine m if the rank of K is less than m
end

V11 = V(1:m,1:m);
[L,p] = chol(sigma2n*eye(m,m)+ V' * V);
if ( p ~= 0 )
   %  necessary if (sigma2n*eye(m,m) + V'*V) is is not SPD (numerically)
   %  reduce rank since (sigma2n*eye(m,m) + V'*V) does not have a nonsing. 
   %  chol factor. Select the 1st m entries of INDEX for the active set.
   m = p-1;
   V = V(:,1:m);
   V11 = V11(1:m,1:m);
end

L = L';                                             % L is now lower triangular

y = y(piv);           % pivot y to be consistent with the pivoting in forming V
V_t_y = V' * y;                                    % used to calculate mu below

% K1star is cov between active set and test points and vstar is variances at
% test points, incl noise variance

INDEX = piv(1:m); % INDEX now identifies the active set determined by chol_part
[vstar, K1star] = feval(covfunc{:}, logtheta, x(INDEX,:), xstar); 
K1star = K1star';        % K1star is now cov between test points and active set 

mu = K1star * ( V11' \ ( L' \ ( L \  V_t_y  )  ) );               
      
if ( nargout >= 3 )
   % calculate variances 
   K1star = K1star / V11' ;                   % overwrite K1star to save memory
   S2PP = vstar - sum(K1star .* K1star, 2 );
   K1star = ( K1star / L' );                  % overwrite K1star to save memory
   S2SR = sigma2n * sum(K1star .* K1star,2); % noise-free SR variance, eq. 8.15
   S2PP = S2PP + S2SR;                 % PP variance eq. (8.27) including noise
   S2SR = S2SR + sigma2n ;                        % SR variance inclusing noise
end
         
if ( nargout == 2 | nargout == 4 )
   % calculate rmse errors for lower rank approximations
   vt = L \ V_t_y ;
   
   B = triu( vt*ones(1,m) );
   rank_increment = 1;         % rank_increment = 1 for rmse at all lower ranks
   %rank_increment = 10;     % rank_increment =10 (say) to save time and memory
   if ( rank_increment == 1 )
       rankv = 1:m;
   else
       % the rmse errors will be calculated only at the ranks in rankv
       rankv = 1:rank_increment:m;
       if (rankv(end) ~= m ), rankv = [rankv, m]; end
       B = B(:,rankv);
   end
   
   if ( nargin == 6 )
      % calculate rmse errors for estimated y, y_est
      V = V *  (  L' \  B ) ;                        % cols of V now have y_est
      sqrt_n = sqrt(n);
      for i = 1:length(rankv)
         rmse(i) = norm(y - V(:,i)) / sqrt_n; 
      end
   end
   
   if ( nargin == 7 )
      % calculate rmse errors for predicted ystar, y_pred
      if ( nargout <= 2 )
         if ( nstar >= m )
            K1star = K1star * ( V11' \ ( L' \  B ) );
         else
            K1star = ( ( K1star / V11' ) / L' ) * B;
         end
      else
         % make use of ealier computations from variance calculation
         K1star = K1star *  B;
      end
      % columns of K1star now have y_pred
      sqrt_nstar = sqrt(nstar);
      for i = 1:length(rankv)
         rmse(i) = norm(ystar - K1star(:,i)) / sqrt_nstar;
      end
   end
   
   if ( nargout == 2 )
      S2SR = rmse;
   end
   
end
